import glob
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from glyconformer import process_files

def roundtriptime(replica, path):
    """
    Function that plots the (average)round trip time for each replica along the progression of time.
    
    Parameters
    ----------
    folder : str
        path to the desired directory 
        
    Returns
    -------
    """
    for i in range(0,replica):
        rtt0 = pd.read_csv("{}/rtt_rep{}.dat".format(path,i), delim_whitespace = True, comment='#')
        rtt0 = rtt0.index.tolist()
        average = round(sum(rtt0)/len(rtt0),2)

        progress = np.arange(0,len(rtt0))
        plt.rcParams['figure.dpi'] = 300
        plt.plot(progress, rtt0, "x")
        plt.axhline(average, linestyle = "--", c = "red")
        plt.xticks([])
        plt.tick_params(axis='both', which='major', labelsize=15)
        plt.ylabel("Duration [ns]", fontsize = 20)
        plt.text(3, 160, "Average rtt: {}".format(average), fontsize = 20)
        plt.title("Replica {}".format(i), fontsize = 20)
        plt.ylim(-10,200)
        plt.show()
    
def free_energy(path):
    """
    Function that plots the one-dimensional free energy of a torsion angle.
    
    The progression of the surface over time is plotted in different colours. 
    
    Parameters
    ----------
    path : str
        directory path leading to the final directory containing the fes.dat files  
    
    Returns
    -------
    """
    
    for filen in glob.glob(path):

        temp_dirList = process_files._read_in_folder(filen,'')
        dirList = []
        for folder in temp_dirList:
            dirList.append(filen+folder)
        del temp_dirList

        fig = plt.figure( figsize=(16, 9), facecolor='w')
        plt.subplots_adjust(left=0.16, bottom=0.15, right=0.9, top=None, wspace=None, hspace=0.05)

        counter = 0
        for ii in dirList:

            plt.rcParams['figure.dpi'] = 300
            col = np.loadtxt(ii, unpack=True, usecols=[0,1])
            plot = plt.plot(col[0], col[1],'-', linewidth=2, zorder=2, alpha=1, color = plt.cm.jet(counter/(len(dirList)-1)))

            plt.xlabel(r'Torsion angle [rad]', fontsize=25)
            plt.ylabel(r'Free energy (kJ/mol)', fontsize=25)

            ax1 = plt.gca()
            for tick in ax1.xaxis.get_major_ticks():
                tick.label1.set_fontsize(20)
            for tick in ax1.yaxis.get_major_ticks():
                tick.label1.set_fontsize(20)

            counter += 1

        plt.show()

def twodim_pucker(path, file, shape):
    """
    Function that plots a two-dimensional pucker free energy profile.
    
    Plots are along variables phi and theta of the Cremer-Pople coordinates. 
    Phi is plotted along the x-axis and theta along the y-axis. 
    The function was only test on files produced by modules of PLUMED.
    
    Parameters
    ----------
    path : str
        directory path to the input file
    file : str
        colvar file that contains the data points x,y and z in the first three consecutive columns
    shape : int
        final shape in x and y direction of the dataset , e.g.: [200,201]
    
    Returns
    -------
    """
    
    x,y,z = np.loadtxt("{}/{}".format(path,file), usecols=(0,1,2), unpack=True)

    X = np.reshape(x, (shape[0], shape[1]))
    Y = np.reshape(y, (shape[0], shape[1]))
    Z_unshifted = np.reshape(z, (shape[0], shape[1]))
    Z_minimum = Z_unshifted.min()

    xnew = np.linspace(-np.pi/2., np.pi/2.,shape[1])
    for i in range(len(X)):
        X[i] = xnew

    #Shift minimum to = 0 in FEP
    if Z_minimum > 0:
        Z = Z_unshifted - Z_minimum
    elif Z_minimum == 0:
        print("Data is already shifted to 0 for its minimum value")
        Z = Z_unshifted
    else:
        Z = Z_unshifted + abs(Z_minimum)
    Z_min = Z.min()
    Z_max = Z.max()

    # Plot 2D free energy surface with heatmap and legend 
    plt.rcParams['figure.dpi'] = 600
    fig = plt.figure()

    ax = fig.add_subplot(111, projection='mollweide')
    im = ax.pcolormesh(Y, X, Z, cmap=plt.cm.jet, vmin = 0, vmax = 40)

    ax.plot([0, 0], [-1.04, 1.04], color='black', linewidth = 0.5)
    ax.plot([0.523599, 0.523599], [-1.04, 1.04], color='black', linewidth = 0.5)
    ax.plot([-0.523599, -0.523599], [-1.04, 1.04], color='black', linewidth = 0.5)
    ax.plot([1.04, 1.04], [-1.04, 1.04], color='black', linewidth = 0.5)
    ax.plot([-1.04, -1.04], [-1.04, 1.04], color='black', linewidth = 0.5)
    ax.plot([1.57, 1.57], [-1.04, 1.04], color='black', linewidth = 0.5)
    ax.plot([-1.57, -1.57], [-1.04, 1.04], color='black', linewidth = 0.5)
    ax.plot([2.09, 2.09], [-1.04, 1.04], color='black', linewidth = 0.5)
    ax.plot([-2.09, -2.09], [-1.04, 1.04], color='black', linewidth = 0.5)
    ax.plot([2.61, 2.61], [-1.04, 1.04], color='black', linewidth = 0.5)
    ax.plot([-2.61, -2.61], [-1.04, 1.04], color='black', linewidth = 0.5)

    ax.plot([-np.pi, np.pi], [1.57, 1.57], color='black', linewidth = 0.5)
    ax.plot([-np.pi, np.pi], [1.04, 1.04], color='black', linewidth = 0.5)
    ax.plot([-np.pi, np.pi], [0.523, 0.523], color='black', linewidth = 0.5)
    ax.plot([-np.pi, np.pi], [0, 0], color='black', linewidth = 0.5)
    ax.plot([-np.pi, np.pi], [-0.523, -0.523], color='black', linewidth = 0.5)    
    ax.plot([-np.pi, np.pi], [-1.04, -1.04], color='black', linewidth = 0.5)
    ax.plot([-np.pi, np.pi], [-1.57, -1.57], color='black', linewidth = 0.5)

    cbar = plt.colorbar(im)
    cbar.set_label("kJ/mol", fontsize = 15)

    plt.xticks([])
    plt.yticks([])
    plt.show()

    
def graph_eval(input_dir, features, maxima_dict, cluster_dict, label_dict):
    """
    Function that plots free energy profiles with annotated minima and maxima.
    
    Evaluation of how good the identification of minima and maxima worked.
    
    Parameters
    ----------
    input_dir : str
        directory from which to read the files for the different torsion angle arrays
    features : str
        list of torsion angle names in the correct order
    maxima_dict : dict
        dictionary that stores the maxima for each feature
    cluster_dict : dict
        dictionary that holds the minima for each feature
    label_dict : dict
        dictionary with features as keys and labels for each minima as values 
    
    Returns
    -------
    """

    y0 = 0
    y1 = 40
    C = np.arange(-0.5236, 0 + 0.001, 0.001)
    c = np.arange(0, 0.5236 + 0.001, 0.001)
    g = np.arange(0.5236, 1.5708 + 0.001, 0.001)
    a = np.arange(1.5708, 2.6180 + 0.001, 0.001)
    T = np.arange(2.6180, 3.5 + 0.001, 0.001)
    t = np.arange(-3.5, -2.6180 + 0.001, 0.001)
    A = np.arange(-2.6180, -1.5708 + 0.001, 0.001)
    G = np.arange(-1.5708, -0.5236 + 0.001, 0.001)
    
    co = "darkslategrey"
    
    fig, axs = plt.subplots(len(features), figsize = (6,60), constrained_layout=True, )
    
    for i, f in zip(range(len(features)),features):
        profile = pd.read_csv("{}/fes_{}.dat".format(input_dir, f), delim_whitespace=True, names = ["x","y"])
        axs[i].set_title('{}'.format(f),fontsize = 15)
        axs[i].set_xlabel("Torsion angle [rad]", fontsize = 12)
        axs[i].set_ylabel("Free energy [kJ/mol]", fontsize = 12)
        axs[i].plot(profile.loc[:,"x"], profile.loc[:,"y"], c = "black", linewidth=2)
        axs[i].fill_between(C, y0, y1, alpha = 0.6, color = "paleturquoise", edgecolor = "none")
        axs[i].fill_between(c, y0, y1, alpha = 0.6, color = "paleturquoise", edgecolor = "none")
        axs[i].fill_between(a, y0, y1, alpha = 0.6, color = "darkturquoise", edgecolor = "none")
        axs[i].fill_between(g, y0, y1, alpha = 0.6, color = "aqua", edgecolor = "none")
        axs[i].fill_between(T, y0, y1, alpha = 0.6, color = "teal", edgecolor = "none")
        axs[i].fill_between(t, y0, y1, alpha = 0.6, color = "teal", edgecolor = "none")
        axs[i].fill_between(G, y0, y1, alpha = 0.6, color = "aqua", edgecolor = "none")
        axs[i].fill_between(A, y0, y1, alpha = 0.6, color = "darkturquoise", edgecolor = "none")
        if len(maxima_dict["{}".format(f)]) == 1:
            axs[i].axvline(maxima_dict["{}".format(f)][0], c = co, linestyle = "--")
            axs[i].text(cluster_dict["{}".format(f)][0],30, label_dict["{}".format(f)][0], fontsize = 12, c = co, weight='bold')
        elif len(maxima_dict["{}".format(f)]) == 2:
            axs[i].axvline(maxima_dict["{}".format(f)][0], c = co, linestyle = "--")
            axs[i].axvline(maxima_dict["{}".format(f)][1], c = co, linestyle = "--")
            axs[i].text(cluster_dict["{}".format(f)][0],30, label_dict["{}".format(f)][0], fontsize = 12, c = co, weight='bold')
            axs[i].text(cluster_dict["{}".format(f)][1],30, label_dict["{}".format(f)][1], fontsize = 12, c = co, weight='bold')
            axs[i].text(cluster_dict["{}".format(f)][2],30, label_dict["{}".format(f)][2], fontsize = 12, c = co, weight='bold')
    
        elif len(maxima_dict["{}".format(f)]) == 3:
            axs[i].axvline(maxima_dict["{}".format(f)][0], c = co, linestyle = "--")
            axs[i].axvline(maxima_dict["{}".format(f)][1], c = co, linestyle = "--")
            axs[i].axvline(maxima_dict["{}".format(f)][2], c = co, linestyle = "--")
            axs[i].text(cluster_dict["{}".format(f)][0],30, label_dict["{}".format(f)][0], fontsize = 12, c = co, weight='bold')
            axs[i].text(cluster_dict["{}".format(f)][1],30, label_dict["{}".format(f)][1], fontsize = 12, c = co, weight='bold')
            axs[i].text(cluster_dict["{}".format(f)][2],30, label_dict["{}".format(f)][2], fontsize = 12, c = co, weight='bold')
            axs[i].text(cluster_dict["{}".format(f)][3],30, label_dict["{}".format(f)][3], fontsize = 12, c = co, weight='bold')

    plt.show()
    
def convergence(binary_path,window):
    """
    Function that computes the moving average and plots the probability of the three most populated conformers over time, accessing convergence.
    
    Parameters
    ----------
    binary_path : str
        complete path to binary COLVAR file, including the file
    window : int
        length of the window used for computing the average
    
    Returns
    -------
    """
    
    binary = pd.read_csv(binary_path, delim_whitespace = True)
    binary['Conformer'] = binary[binary.columns[0:len(binary.columns)]].apply(lambda x: ''.join(x.dropna().astype(str)),axis=1)
    binary = binary.drop(binary.columns[0:len(binary.columns)-1], axis=1)
    
    count = binary.groupby('Conformer').size().to_frame(name = 'count').reset_index()
    count = count.sort_values(by=['count'], ascending=False, ignore_index = True)
    top = [count.loc[0, "Conformer"], count.loc[1, "Conformer"], count.loc[2, "Conformer"]]
    
    label = ["1st","2nd","3rd"]
    color = ["#173c4d","#146b65", "#4e9973"]
    
    plt.rcParams['figure.dpi'] = 600
    plt.figure()
    plt.xlabel("Time [µs]", fontsize=15)
    plt.ylabel("Probability [%]", fontsize=15)
    plt.tick_params(axis='both', which='major', labelsize=15)
    
    plt.ylim(0,100)
    
    for n in (range(0,3)):
        
        binary = pd.read_csv(binary_path, delim_whitespace = True)
        binary['Conformer'] = binary[binary.columns[0:len(binary.columns)]].apply(lambda x: ''.join(x.dropna().astype(str)),axis=1)
        binary = binary.drop(binary.columns[0:len(binary.columns)-1], axis=1)
        
        for i in binary.index:
            if binary.loc[i,"Conformer"] == top[n]:
                pass
            else:
                binary.loc[i,"Conformer"] = np.nan

        binary["rollcount"] = binary.rolling(window, min_periods=1).count()
        binary.loc[:, "rollcount"] = [x / window for x in binary["rollcount"]]
        binary.loc[:, "rollcount"] = [x * 100 for x in binary["rollcount"]]
        plt.plot(binary.index/125, binary.loc[:, "rollcount"], label = label[n], c = color[n], linestyle = "-")
        
    plt.legend(fontsize = 15)
    plt.show()



"""
This module contains functions to process free energy
files produced by the PLUMED sum_hills command.

List of functions
rename_profiles
shift_to_zero
"""
import glob
import os

def _read_in_folder(folder, datatype):
    """
    Helper function that constructs a list of all files present in a directory
    
    Parameters
    ----------
    folder : str
        Path to the desired directory from which files
        should be checked out
    
    Returns
    -------
    dirList : str
        List containing path+file names
    """
    dirList = []
    for files in os.listdir(folder):
        if files.endswith(datatype):
            dirList.append(files)
    dirList.sort()
    if dirList == []:
        return None
    else:
        return dirList

def _get_CV(LINES):
    """
    Read the collective variable column of the .dat file
    
    Parameters
    ----------
    LINES : float
        Lines read in from a file
    
    Returns
    -------
    CV : float
        List of collective variables
    """
    CV = []
    for line in LINES:
        words = line.split()
        CV.append(float(words[0]))
    return CV

def _get_FE(LINES):
    """
    Read the free energy column of the .dat file
    
    Parameters
    ----------
    LINES : float
        Lines read in from a file
    
    Returns
    -------
    FE : float
        List of free energy values along the collective variable
    """
    FE = []
    for line in LINES:
        words = line.split()
        FE.append(float(words[1]))
    return FE

def _get_min(FILES):
    """
    Searches for the minimum value in the free energy columns 
    
    Parameters
    ----------
    FILES : str
        list of path+files that should be checked for the 
        minimum energy value
    
    Returns
    -------
    MIN : float
        lowest free energy value
    """
    FILE = open(FILES[-1], "r")
    LINES = FILE.readlines()[5:]
    FE = _get_FE(LINES)
    MIN = min(FE)
    return MIN


def _shift_FE(FE, MIN):
    """
    Shifts the energy column by the minimum free energy to align it to 0
    
    Parameters
    ----------
    FE : float
        List of free energy values along the collective variable
    MIN : float
        lowest free energy value
    
    Returns
    -------
    FE_NEW : float
        List of updated free energy values along the collective variable
    """
    FE_NEW = []
    for fe in FE:
        fe_shifted = fe - MIN
        FE_NEW.append(fe_shifted)
    return FE_NEW

    
def rename_profiles(path):
    """
    Function that reformats fes file names in order to get them read sorted.
    
    Instead of label like *1.dat, *2.dat ..., the code converts it 
    to *00001.dat, *00002.dat and so on.
    
    Parameters
    ----------
    path : str
        directory path including the folder with the fes files
    
    Returns
    -------
    """
    temp_dirList = _read_in_folder(path,'')
    dirList = []
    for folder in temp_dirList:
        dirList.append(path+folder)
    del temp_dirList
    for ii in range(len(dirList)):
        os.system('mv %sfes_%d.dat %sfes_%05d.dat' %(path, ii, path, ii))

def shift_to_zero(path):
    """
    Function that shifts all fes profiles in the path to zero.
    
    In order to get the free energy profiles aligned to zero,
    the code checks for the minimum free energy value and substracts
    it from all free energy profiles.
    
    Parameters
    ----------
    path : str
        directory path including the folder with the fes files
    
    Returns
    -------
    """
    FILES = glob.glob(path + "/fes*.dat")
    FILES.sort()
    MIN = _get_min(FILES)

    for item in FILES:
        FILE = open(item, "r")
        LINES = FILE.readlines()[5:]
        FILE.close()
        CV = _get_CV(LINES)
        FE = _get_FE(LINES)
        FE_NEW = _shift_FE(FE, MIN)
        OUTPUT = open(item, "w")
        for num in range(len(CV)):
            OUTPUT.write("{:<14.9F}{:<14.9F}\n".format(CV[num], FE_NEW[num]))
        OUTPUT.close()



"""
This module contains functions to generate N-glycan
conformer strings from a COLVAR file containing
torsion angle values recorded over a MD simulation.

List of functions:
conformer_limit_string
create_binary
find_min_max
label_min
load_dict
perform_block_averages
plot_distribution
safe_dict
"""
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import plumed
from scipy.signal import argrelextrema
import json
import matplotlib.ticker as ticker

def _combine_dict(d1, d2, d3, d4):
    """
    Function that merges values from different 
    dictionaries having the same key into one.
    
    Parameters
    ----------
    d1,d2,d3,d4 : dict (hash table type)
        Up to 4 dictionaries 
    
    Returns
    -------
    dict
        1 merged dictionary
    """
    
    return {
        k: tuple(d[k] for d in (d1, d2, d3, d4) if k in d)
        for k in set(d1.keys()) | set(d2.keys()) | set(d3.keys()) | set(d4.keys())
    }

def _vertical_conformer_string(namelist):
    """
    Function that converts the conform string from a 
    horizontal format to a vertical one.
    
    Parameters
    ----------
    namelist : str
        List of conformer strings 
        
    Returns
    -------
    namelist_v : str
        Conformer strings as list in a vertical format
    """
    
    namelist_v = []
    for i in range(len(namelist)):
        res = ''
        for j,ele in enumerate(namelist[i]):
            if j in (range(2,1000,3)):
                res += ele + "\n"
            else:
                res += ele
        namelist_v += [res]
    return namelist_v

def _include_branch_conformer_string(namelist, branches, features):
    """
    Function that inserts branch separators and replaces 
    labels by dots in the conform strings.
    
    Parameters
    ----------
    namelist : str
        List of conformer strings 
    branches : str
        List of branch names present in the glycan
    features : str
        List of features
        
    Returns
    -------
    namelist : str
        List of conformer strings with branch separators 
        and repetative labels exchange by dots
    """
    
    for i in range(1,len(namelist)):
        
        for j,k in zip(range(0,len(features)*3,3),range(3,len(features)*3+1,3)):
            if namelist[i][j:k] == namelist[0][j:k]:
                newname = namelist[i][:j] + " • " + namelist[i][k:]
                namelist[i] = newname
            else: 
                pass 
        # Limitation: Can only replace branch separators which do occur max. twice
        for b in branches:
            index = namelist[0].find(b)
            newbranch = namelist[i][:index] + "───" + namelist[i][index+3:]
            namelist[i] = newbranch
        
        for b in branches:
            rindex = namelist[0].rfind(b)
            if rindex == "-1":
                pass
            else:
                rnewbranch = namelist[i][:rindex] + "───" + namelist[i][rindex+3:]
                namelist[i] = rnewbranch

    return namelist

def safe_dict(dict_name, filename):   
    """
    Function saving a dictionary to file.
    
    Parameters
    ----------
    dict_name : dict
        Name of the dictionary to safe
    filename : str
        Path + name of the file with file extension     
    
    Returns
    -------
    """
    
    with open('{}'.format(filename), 'w') as file:
        file.write(json.dumps(dict_name))
        
def load_dict(filename):
    """
    Function loading a dictionary from file.
    
    Parameters
    ----------
    filename : str
        Name of the path + file to read in,
        including file extension 
    
    Returns
    -------
    dict
        Variable holding the dictionary
    """
    
    with open('{}'.format(filename)) as f:
        data = f.read()
        print(data)
    dict = json.loads(data)
    
    return dict

def find_min_max(input_dir, features, order_max, order_min):
    """ 
    Finds minima and maxima of a 1D array read from file.
    
    Function reading in the free energy profiles (as arrays 
    from fes_*.dat files) of each feature (torsion angle) and 
    determining the maxima and minima of the array. The minima 
    array is altered in a way that an array with 2 maxima 
    corresponds to 3 minima, considering periodicity.
    
    Parameters
    ----------
    features : str
        List of torsion angle names (= column names in the 
        COLVAR file) in the correct order
    input_dir : str
        Directory from which to read the free energy files 
        (fes_*.dat) for the different torsion angles
    order_max: int
        How many points on each side of a maxima to use for 
        the comparison to consider comparator(n, n+x) to be True.
    order_min: int
        How many points on each side of a minima to use for 
        the comparison to consider comparator(n, n+x) to be True.
    
    Raises
    ------
    ValueError
        If free energy has more than 4 minimas, the algorithm can
        not handle it anymore.
    
    Returns
    -------
    maxmima_dict : dict
        Dictionary with features as keys and maxima points as values
    minima_dict : dict
        Dictionary with features as keys and minima points as values
    """

    maxima_dict = {}
    minima_dict = {}

    for f in features:
        profile = pd.read_csv("{}/fes_{}.dat".format(input_dir, f), delim_whitespace=True, names = ["x","y"])
        profmin = profile.index[profile['y'] == 0.0]
        profmin_value = profile.iloc[profmin[0],0]
        p1 = profile.iloc[:profmin[0],:]
        p2 = profile.iloc[profmin[0]:,:]
        p = p2.append([p1])
        p = p.to_numpy()
        
        maxima = argrelextrema(p[:,1], np.greater, order=order_max)
        minima = argrelextrema(p[:,1], np.less, order=order_min)
        
        if len(maxima[0]) == 0:
            maxima_dict["{}".format(f)] = [0]
        elif len(maxima[0]) == 1:
            maxima_dict["{}".format(f)] = [p[maxima[0][0],[0]].item()]
        elif len(maxima[0]) == 2:
            x = [p[maxima[0][0],[0]].item(), p[maxima[0][1],[0]].item()]
            x.sort()
            maxima_dict["{}".format(f)] = x
        elif len(maxima[0]) == 3:
            x = [p[maxima[0][0],[0]].item(), p[maxima[0][1],[0]].item(), p[maxima[0][2],[0]].item()]
            x.sort()
            maxima_dict["{}".format(f)] = x

            
            
        if len(minima[0]) == 0:
            minima_dict["{}".format(f)] = [profmin_value]
           
        elif len(minima[0]) == 1:   
            x = [profmin_value, p[minima[0][0],[0]].item()]
            x.sort()

            test = -3.5 <= x[0] <= maxima_dict["{}".format(f)][0]
            
            if test == True:                                  
                x = [x[0],x[1],x[0]] 
                minima_dict["{}".format(f)] = x
                        
            else:
                x = [x[1],x[0],x[1]] 
                minima_dict["{}".format(f)] = x    
            
            
        elif len(minima[0]) == 2:
            x = [profmin_value, p[minima[0][0],[0]].item(), p[minima[0][1],[0]].item()]
            x.sort()
            
            test = -3.5 <= x[0] <= maxima_dict["{}".format(f)][0]
         
            if test == True:                                  
                x = [x[0],x[1],x[2],x[0]] 
                minima_dict["{}".format(f)] = x
                        
            else:
                x = [x[2],x[0],x[1],x[2]] 
                minima_dict["{}".format(f)] = x                     
        
        elif len(minima[0]) >= 3:
            raise ValueError('More than 3 minima detected')
    
    return maxima_dict, minima_dict

def label_min(cd, features, f_omega):
    """
    Function that labels minima by their corresponding IUPAC name.
    
    Reads in the created minima_dict, containing the position of 
    minimas along the torsion angles. These values are replaced by 
    capital letters originating from the IUPAC nomenclature for 
    torsion angles. 
    
    Parameters
    ----------
    cd : dict
        Minima_dict
    features : str
        List of torsion angle names (= column names in COLVAR file) 
        in the correct order
    f_omega : str
        List of omega torsion angle names in the correct order
    
    Returns
    -------
    label_dict : dict
        Dictionary with feature names as keys and sorted labels for each minima as values
    """
    
    tangle_dict = {
    " C ": (-0.5236, 0),
    " c ": (0, 0.5236),
    " G₊": (0.5236, 1.5708),
    " A₊": (1.5708, 2.6180),
    " T ": (2.6180, 3.5),
    " t ": (-3.5, -2.6180),
    " A₋": (-2.6180, -1.5708),
    " G₋": (-1.5708, -0.5236)
    }
    
    oangle_dict = {
    " gg": (-2.6180, 0),
    " gt": (0, 2.6180),
    " tg": (-3.5, -2.6180),
    " TG": (2.6180, 3.5)
    }
    
    name_dict = {}
    name_dict1 = {}
    name_dict2 = {}
    name_dict3 = {}


    for f in features:
        if f in f_omega:
            for key, value in oangle_dict.items():
                if value[0] <= cd["{}".format(f)][0] < value[1]:
                    name_dict["{}".format(f)] = key
                if cd["{}".format(f)][0] == 4:
                    name_dict["{}".format(f)] = 'none'      
        else:
            for key, value in tangle_dict.items():
                if value[0] <= cd["{}".format(f)][0] < value[1]:
                    name_dict["{}".format(f)] = key
                if cd["{}".format(f)][0] == 4:
                    name_dict["{}".format(f)] = 'none'

    for f in features:
        if len(cd["{}".format(f)]) > 1:
            if f in f_omega:
                for key, value in oangle_dict.items():
                    if value[0] <= cd["{}".format(f)][1] < value[1]:                    
                        name_dict1["{}".format(f)] = key
                    if cd["{}".format(f)][1] == 4:
                        name_dict1["{}".format(f)] = 'none'        
            else:
                for key,value in tangle_dict.items():
                    if value[0] <= cd["{}".format(f)][1] < value[1]:                    
                        name_dict1["{}".format(f)] = key
                    if cd["{}".format(f)][1] == 4:
                        name_dict1["{}".format(f)] = 'none'

    for f in features:                
        if len(cd["{}".format(f)]) > 2:
            if f in f_omega:
                for key, value in oangle_dict.items():
                    if value[0] <= cd["{}".format(f)][2] < value[1]:                    
                        name_dict2["{}".format(f)] = key
                    if cd["{}".format(f)][2] == 4:
                        name_dict2["{}".format(f)] = 'none'                    
            else:
                for key,value in tangle_dict.items():
                    if value[0] <= cd["{}".format(f)][2] < value[1]:                    
                        name_dict2["{}".format(f)] = key
                    if cd["{}".format(f)][2] == 4:
                        name_dict2["{}".format(f)] = 'none'

    for f in features:                
        if len(cd["{}".format(f)]) > 3:
            if f in f_omega:
                for key, value in oangle_dict.items():
                    if value[0] <= cd["{}".format(f)][3] < value[1]:                    
                        name_dict3["{}".format(f)] = key
            else:
                for key,value in tangle_dict.items():
                    if value[0] <= cd["{}".format(f)][3] < value[1]:                    
                        name_dict3["{}".format(f)] = key

    label_dict = _combine_dict(name_dict,name_dict1,name_dict2,name_dict3)
    return label_dict

def create_binary(maxima_dict, label_dict, colvar_dir, input_dir, 
                  colvar_length, features,* , loc1, loc2, loc3 = 0, 
                  loc4 = 0, loc5 = 0, loc6 = 0, locF = 0, 
                  locG = 0, bra3 = "───", bra4 = "───", bra5 = "───", 
                  bra6 = "───"):
    """
    Converts torsion angle values read from COVLAR file into IUPAC letters.
    
    Function that reads in a COLVAR file with torsion angle values stored 
    in each column and replaces each value by the corresponding IUPAC label, 
    which was previously defined in the label_dict.
    
    Parameters
    ----------
    maxima_dict : dict
        Dictionary that contains the maxima for each feature
    label_dict : dict
        Dictionary with features as keys and labels for each minima as values 
    colvar_dir : str
        Path to the COLVAR file with stored features as columns
    colvar_length : int
        Desired length of the COLVAR file
    features : str
        List of torsion angle names (= column names) in the correct order
    glycantype : str
        Define type of glycan from "complex", "mannose", or "hybride"
    loc3,loc4,loc5,loc6 : int 
        Location where to put the branch in the confromer string. Indicate
        the positon by an integer number
     bra3,bra4,bra5,bra6 : str
        Name of the separator to be used for the branch, e.g. "2──"
        
    Returns
    -------
    c : str
        Original COLVAR input as a pandas dataframe in which the torsion 
        angle values are replaced by labels
    ccf : mixed
        Pandas dataframe with a column for the conformer string and another 
        column for the occurance of the conformer (counts)
    features : str
        Updated feature list with branch separators
    branches : str
        List of branches present in the processed glycan structure
    """
   
    colvar = plumed.read_as_pandas(colvar_dir)
    colvar = colvar[features]
    colvar = colvar.iloc[::round(colvar.shape[0]/colvar_length), :]

    c = plumed.read_as_pandas(colvar_dir)
    c = colvar[features]
    c = colvar.iloc[::round(colvar.shape[0]/colvar_length), :]

    for f in features:
        if len(maxima_dict["{}".format(f)]) == 1:

            c["{}".format(f)] = np.where((-3.5 <= colvar["{}".format(f)]) & (colvar["{}".format(f)] <= 3.5) , label_dict["{}".format(f)][0], c["{}".format(f)])
            
 
        elif len(maxima_dict["{}".format(f)]) == 2:

            c["{}".format(f)] = np.where((-3.5 <= colvar["{}".format(f)]) & (colvar["{}".format(f)] < maxima_dict["{}".format(f)][0]) , label_dict["{}".format(f)][0], c["{}".format(f)])
            c["{}".format(f)] = np.where((maxima_dict["{}".format(f)][0] <= colvar["{}".format(f)]) & (colvar["{}".format(f)] < maxima_dict["{}".format(f)][1]) , label_dict["{}".format(f)][1], c["{}".format(f)])
            c["{}".format(f)] = np.where((maxima_dict["{}".format(f)][1] <= colvar["{}".format(f)]) & (colvar["{}".format(f)] <= 3.5) , label_dict["{}".format(f)][2], c["{}".format(f)])      
            
        elif len(maxima_dict["{}".format(f)]) == 3:

            c["{}".format(f)] = np.where((-3.5 <= colvar["{}".format(f)]) & (colvar["{}".format(f)] < maxima_dict["{}".format(f)][0]) , label_dict["{}".format(f)][0], c["{}".format(f)])
            c["{}".format(f)] = np.where((maxima_dict["{}".format(f)][0] <= colvar["{}".format(f)]) & (colvar["{}".format(f)] < maxima_dict["{}".format(f)][1]) , label_dict["{}".format(f)][1], c["{}".format(f)])
            c["{}".format(f)] = np.where((maxima_dict["{}".format(f)][1] <= colvar["{}".format(f)]) & (colvar["{}".format(f)] < maxima_dict["{}".format(f)][2]) , label_dict["{}".format(f)][2], c["{}".format(f)])
            c["{}".format(f)] = np.where((maxima_dict["{}".format(f)][2] <= colvar["{}".format(f)]) & (colvar["{}".format(f)] <= 3.5) , label_dict["{}".format(f)][3], c["{}".format(f)])
   
            
    c = pd.DataFrame(c, columns = c.columns)
    c = c.replace(' t ', ' T ')
    c = c.replace(' TG', ' tg')
    
    branches = []
    c.insert(loc=loc1, column='sep1', value="6──")
    branches.append("6──")
    c.insert(loc=loc2+1, column='sep2', value="3──")
    branches.append("3──")
    
    if loc3 == 0:
        pass
    else:
        c.insert(loc=loc3, column='sep3', value=bra3)
        branches.append(bra3)
    
    if loc4 == 0:
        pass
    else:
        if loc3 != 0:
            c.insert(loc=loc4, column='sep4', value=bra4)
            branches.append(bra4)
        else:
            raise ValueError("loc3 need to be defined before loc4")
            
    if loc5 == 0:
        pass
    else:
        if loc4 != 0:
            c.insert(loc=loc5, column='sep5', value=bra5)
            branches.append(bra5)
        else:
            raise ValueError("loc4 need to be defined before loc5")
    
    if loc6 == 0:
        pass
    else:
        if loc6 != 0:
            c.insert(loc=loc6, column='sep6', value=bra6)
            branches.append(bra6)
        else:
            raise ValueError("loc5 need to be defined before loc6")
    
    if locF == 0:
        pass
    else:
        c.insert(loc=locF, column='sepF', value="f──")
        branches.append("f──")
        
    if locG == 0:
        pass
    else:
        c.insert(loc=locG, column='sepG', value="g──")
        branches.append("g──")
        
    features = list(c.columns)
    ccf = c.groupby(features).size().to_frame(name = 'count').reset_index()
    ccf['Conformer'] = ccf[ccf.columns[0:len(features)]].apply(lambda x: ''.join(x.dropna().astype(str)),axis=1)
    ccf = ccf.drop(c.columns[0:len(features)], axis=1)
    c.to_csv("{}/COLVAR_binary".format(input_dir) ,sep = " ", index = False)
    
    return c, ccf, features, branches
    
def perform_block_averages(c, ccf, store_dir):
    """
    Calculates probability of conformers for 10 separate blocks.
    
    Function that splits the torsion angle COLVAR dataset in 10 
    blocks to do the conformer labelling and probability calculation 
    for each block.
    
    Parameters
    ---------- 
    c : str
        Binary COLVAR input file as a pandas dataframe in which the 
        torsion angle values are replaced by labels
    ccf : mixed
        Pandas dataframe with the first column for the conformer string 
        and the second column for the occurance of the conformer (counts)
    store_dir : str
        Directory name to store the conformer files in
        
    Returns
    -------
    """
    features = list(c.columns)
    ccf = ccf.set_index('Conformer')
    step = len(c)/10

    clear1 = np.arange(0,len(c), step)
    clear2 = np.arange(step, len(c) + step, step)
    clear1 = list(map(int, clear1))
    clear2 = list(map(int, clear2))
    
    for i,j,k in zip(range(1,11), clear1, clear2):
        c_short = c.iloc[j:k,:]
        cc_short = c_short.groupby(features).size().to_frame(name = 'count').reset_index()
        norm = len(c_short)
        p = (cc_short.iloc[:,len(features)] / norm)
        cc_short["Probability"] = p
        cc_short['Conformer'] = cc_short[cc_short.columns[0:len(features)]].apply(lambda x: ''.join(x.dropna().astype(str)),axis=1)
        cc_short = cc_short.drop(cc_short.columns[0:len(features)], axis=1)
        # Join both dataframes taking 'Conformer' as key column for comparison
        cc_short = cc_short.set_index('Conformer')

        ccc_short = cc_short.join(ccf,how='outer',lsuffix='_partial', rsuffix='_full')
        ccc_short = ccc_short.round(6)
        ccc_short = ccc_short.fillna(0.0)
        ccc_short = ccc_short.reset_index()

        # Print partially data sorted to file
        ccc_short.to_csv("{}/Cluster_conformer{}.dat".format(store_dir,i), sep= " ", header=False)

def plot_distribution(input_dir, limit, branches, features):
    """
    Plots a bar graph of the conformer density function.
    
    Function that plot the probability distribution of the 
    most occuring conformers. The limit defines the cutoff
    probability, where conformers are excluded from the 
    representation if below the cutoff.
    
    Parameters
    ----------
    input_dir: str
        Path to directory in which the "Cluster_conformer" 
        files from the block averaging step are stores
    limit: float
        Propability limit up to which conformers are included in the graph 
    branches: str
        List of branches present in the glycan
    features: str
        Updated features list
        
    Returns
    -------
    """

    hist1 = pd.read_csv("{}/Cluster_conformer1.dat".format(input_dir), delim_whitespace=True, names = ["Index", "Conformer", "Count_partial", "Prob", "Count_full"] )
    N, average, average2 = 1, hist1.iloc[:,3], hist1.iloc[:,3]*hist1.iloc[:,3]
    for i in range(2,11): 
        histn = pd.read_csv("{}/Cluster_conformer{}.dat".format(input_dir, i), delim_whitespace=True, names = ["Index", "Conformer", "Count_partial", "Prob", "Count_full"])
        N, average, average2 = N + 1, average + histn.iloc[:,3], average2 + histn.iloc[:,3]*histn.iloc[:,3]
    

    average = pd.DataFrame(average/N, columns=['Prob'])
    average = average.drop(average[average['Prob'] < limit ].index)
    average = average.sort_values(by='Prob',ascending=False)
    indexlist = average.index.tolist()

    hist = pd.read_csv("{}/Cluster_conformer1.dat".format(input_dir), names = ["Index", "Conformer", "Count_partial", "Prob", "Count_full"], sep = " ", dtype = str)
    hist = hist.filter(indexlist, axis = 0)
    namelist = hist['Conformer'].tolist()
    name_list = _include_branch_conformer_string(namelist, branches, features)
    name_list = _vertical_conformer_string(namelist)
    average2 = average2.to_frame()
    average2.columns = ["Error"]
    average2 = average2.filter(indexlist, axis = 0)

    # Final variances
    var = (N/(N-1))*( average2.Error / N - average.Prob*average.Prob ) 
    # Errors
    error = np.sqrt( var / N )
 
    pos_list = np.arange(len(name_list))
    
    colors = ["#173c4d","#146b65","#4e9973","#a7c09f","#dfa790","#c76156","#9a2b4c","#600b4a"]
    plt.rcParams['figure.dpi'] = 300
    ax = plt.axes()
    ax.xaxis.set_major_locator(ticker.FixedLocator((pos_list)))
    ax.xaxis.set_major_formatter(ticker.FixedFormatter((name_list)))   
    bar = plt.bar(pos_list, average.Prob * 100, color = "blue", linewidth = "1", yerr = error * 100)
    for i in range(len(pos_list)):
        bar[i].set_color(colors[i])
    
    plt.xlabel("Conformer")
    plt.ylabel('Probability [%]')
    plt.savefig("{}/Conformer_population_full.png".format(input_dir))

    
def conformer_limit_string(limit, directory):
    
    """
    Prints out the conform strings which are occuring more or less than the limit.
    
    Reads in constructed conformer probability records (Cluster_conformer*d.dat files) 
    from 10 blocks and averages them to sort conformers by their occurance. The limit 
    sets which conformers are discarded (below) and which are included for analysis (above).
    
    Parameters
    ----------
    limit: float
        Propability limit up to which conformers are excluded from coloured representation in DimRed
    directory: str
        Name of directory path from which to read the Cluster_conformer.dat files
        
    Returns
    -------
    targets_below: str
        List of conformer strings below the limit
    targets_above: str
        List of conformer strings above the limit
    """

    # Load partially data sorted 
    hist1 = pd.read_csv("{}/Cluster_conformer1.dat".format(directory), delim_whitespace=True, names = ["Index","Conformer","Count", "Prob", "Count_all"])
    N, average = 1, hist1.iloc[:,3]
    for i in range(2,11): 
        histn = pd.read_csv("{}/Cluster_conformer{}.dat".format(directory, i), delim_whitespace=True, names = ["Index","Conformer","Count", "Prob", "Count_all"])
        N, average = N + 1, average + histn.iloc[:,3]

    # Sort conformers by probability
    average = average.to_frame()
    average = average.div(N)
     
    # below
    below = average.drop(average[average['Prob'] > limit ].index, inplace=False)
    below = below.sort_values(by='Prob',ascending=False)
    indexlist = below.index.tolist()
    hist = pd.read_csv("{}/Cluster_conformer1.dat".format(directory), sep = " ", names = ["Index","Conformer","Count", "Prob", "Count_all"], dtype = str)
    hist = hist.filter(indexlist, axis = 0)
    targets_below = hist['Conformer'].tolist()

    # above
    above = average.drop(average[average['Prob'] < limit ].index, inplace=False)
    above = above.sort_values(by='Prob',ascending=False)
    indexlist = above.index.tolist()
    # Safe to use no.1 or any other of these files. All have all conformers included
    hist = pd.read_csv("{}/Cluster_conformer1.dat".format(directory), sep = " ", names = ["Index","Conformer","Count", "Prob", "Count_all"], dtype = str)
    hist = hist.filter(indexlist, axis = 0)
    targets_above = hist['Conformer'].tolist()

    return targets_below,targets_above



from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()
classifiers = [
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Education',
    'Programming Language :: Python :: 3'
    ]

setup(name='GlyCONFORMER',
      version='0.0.1',
      description="GlyCONFORMERS is a Python package that assigns conformer strings to N-glycan conformers, based on their torsion angle values.",
      url="https://github.com/IsabellGrothaus/GlyCONFORMER",
      long_description = (here / "README.rst").read_text(encoding="utf-8"),
      author='Isabell Grothaus',
      author_email='grothaus@uni-bremen.de',
      classifiers=classifiers,
      keywords="glycan, conformer, classification, sugar, carbohydrates",
      license='GPL-3.0 license',
      packages=find_packages(),
      install_requires=[
                       "numpy","matplotlib","pandas","plumed","scipy"
                       ],
      )
